/*
 *Detemine CPU type and features using cpuid X86 instruction
 *Contributed by R. Lunnon
 *Free for public use
 *
 */
#include "cpuid.h"
#include <stdlib.h>

#include <stdio.h>
#include <string.h>


char *cpu_features[] ={
"FPU Supported"
,"Virtual Mode Extensions"
,"Breakpoint functions"
,"Page Size Extensions"
,"Time Stamping"
,"Model Specific registerrs"
,"Page Address Extensions"
,"Machine Check Exception"
,"Compare exchange 8 byte instruction"
,"APIC present"
,"CPU Feature 10"
,"Fast system call extensions"
,"Memory Type/range regs supported"
,"Global paging register"
,"Machine Check Architecture"
,"Conditional move instructions"
,"Page Atribute table"
,"Page address extensions to 36 bits"
,"Processor Serial Number (INTEL)"
,"CLFLUSH Instruction (INTEL)"
,"CPU_FEATURE_20"
,"Debug Store Feature"
,"AMD MMX Extensions / ACPI"
,"INTEL MMX"
,"Fast Floating Point Save/SSE"
,"Intel SIMD"
,"Intel SIMD 2"
,"Self Snoop"
,"CPU_FEATURE_28"
,"Thermal Monitor"
,"AMD 3D Now Extensions"
,"AMD 3D Now"
,NULL};


#ifdef CPUIDDEBUG
static char *PrintFeatures(unsigned long descriptor,char *buffer)
{

	int i;
	char *loc = buffer;
	for(i=0;i<32;i++)
	{
		if((descriptor>>i) &1 )
			loc+=sprintf(loc,"%s\n",cpu_features[i]);
	}
	return (buffer);
}
#endif
static cpuid_t CpuId(unsigned long inst)
{

	static unsigned long _eax,_ebx,_ecx,_edx;
	static cpuid_t cpuid;


	_eax=inst;
	_ebx=_edx=_ecx=0;


/*
 * Note this is supposed to work on 386s
 * It is not very likely wine will be in use on these systems
 * Note that this method will fail if an interrupt modifies the ID bit
 * of the flags register, but I expect this is unlikely to happen in practise
 * If it does, the worst that can happen is we get identified as a 386 system
 */

	asm (
	"\tpushl %%ebx\n"
	"\tmovl %%eax,%%ecx\n"
	"\tpushfl\n"
	"\tpopl %%eax\n"
	"\tmovl %%eax,%%ebx\n"
	"\txorl  $2097152, %%eax\n"
	"\tpushl %%eax\n"
	"\tpopfl\n"
	"\tpushfl\n"
	"\tpopl %%eax\n"
	"\tpushl %%ebx\n"
	"\tpopfl\n"
	"\tandl $2097152,%%eax\n"
	"\tandl $2097152,%%ebx\n"
	"\txorl %%eax,%%ebx\n"
	"\tjnz have_cpuid\n"
	"\tclrl %%ebx\n"
	"\tjmp endit\n"
	"have_cpuid:\tmovl %%ecx,%%eax\n"
	"\tcpuid\n"
	"endit:\n"
	"\tmovl %%ebx,%0\n"
	"\tpopl %%ebx\n"
	:
	"=r"(_ebx),"=a"(_eax),"=c"(_ecx),"=d"(_edx):"a"(inst)

	);

	cpuid.eax=_eax;
	cpuid.ebx=_ebx;
	cpuid.ecx=_ecx;
	cpuid.edx=_edx;
	*(unsigned long *) &cpuid.IdString[0]=_ebx;
	*(unsigned long *) &cpuid.IdString[8]=_ecx;
	*(unsigned long *) &cpuid.IdString[4]=_edx;

	cpuid.IdString[12]=0;

	return(cpuid);
}


/************************************************
 *  Interpet the Results of the CPUID Instruction
*/

cpuid_t CPUID_GetId()
{
	cpuid_t cpuid,cpuid1,cpuid2;
#ifdef CPUIDDEBUG
	char buffer[ 8000 ];
#endif
	int amd;

	cpuid=CpuId(0);
	if(!strcmp(cpuid.IdString,"AuthenticAMD"))
	{
		cpuid.vendor=amd=CPU_VENDOR_AMD;
		cpuid2=CpuId(0x80000001);
	}
	cpuid1=CpuId(0x00000001);

	if(!cpuid.IdString[0]) {

  		cpuid.nocpuid=1;
		cpuid.features=0;
		cpuid.ex_features=0;
		cpuid.brand=0;
		cpuid.apic=0;
		cpuid.ex_family=0;
		cpuid.ex_model=0;
		cpuid.cpu_type=0;
		cpuid.family=3;
		cpuid.model=0;
		cpuid.stepping=0;
		strcpy(cpuid.IdString,"i386");
	}


/* Lets fill out the rest of the data structure
 * Opcode 1 (in CPUID 1) holds the processor info
 *
 * Store the feature information.
 */
	cpuid.nocpuid=0;
	cpuid.features=cpuid1.edx;
	cpuid.ex_features=cpuid2.edx;
	cpuid.brand=cpuid1.ebx&0xff;
	cpuid.apic=(cpuid1.ebx>>24) &0xff;
	cpuid.ex_family=(cpuid1.eax>>20) &0xff;
	cpuid.ex_model=(cpuid1.eax>>16) &0x0f;
	cpuid.cpu_type=(cpuid1.eax>>12) &0x03;
	cpuid.family=(cpuid1.eax>>8) &0x0f;
	cpuid.model=(cpuid1.eax>>4) &0x0f;
	cpuid.stepping=(cpuid1.eax) &0x0f;

#ifdef CPUIDDEBUG
	printf(" Processor is %s, Ver %x\nFamily %d model %d Step %d\n brand %d apic %d, ex_fam %d ex_model %d type %d \n\n",
			cpuid.IdString,cpuid1.eax,cpuid.family,cpuid.model,cpuid.stepping,cpuid.brand,cpuid.apic,cpuid.ex_family,cpuid.ex_model,cpuid.cpu_type);

	printf("Featuring\n");
	PrintFeatures(amd ? cpuid2.edx :cpuid1.edx, buffer);
	printf(buffer);

#endif
	return(cpuid);
}

int CPUID_IsVendor(int vendor)
{
	cpuid_t cpuid;

	cpuid=CPUID_GetId();
	return(cpuid.vendor==vendor);
}


int CPUID_IsFeatureAvailable(unsigned long feature){
cpuid_t cpuid;

	cpuid=CPUID_GetId();
	if(cpuid.vendor==CPU_VENDOR_AMD) return(!!(cpuid.ex_features&feature));
	else return(!!(cpuid.features&feature));
}



/*************************************************
 * A Simple Main - compile with -DCPUIDDEBUG for a standalone CPUID Test
 */

#ifdef CPUIDDEBUG
main(){
CPUID_GetId();
}
#endif
